%{ open Ast %}

%token SEMI LPAREN RPAREN LBRACE RBRACE LBRACKET RBRACKET COMMA COLON
%token PLUS MINUS TIMES DIVIDE MOD POWER ASSIGN
%token EQ NEQ LT LEQ GT GEQ
%token SQRT SUM AVG ABS LOG
%token RETURN IF ELSE FOR WHILE INT VOID ARRAY FUNC PRINT
%token <int> LITERAL
%token <string> ID
%token EOF

%nonassoc NOELSE
%nonassoc ELSE
%right ASSIGN
%left EQ NEQ
%left LT GT LEQ GEQ
%left PLUS MINUS
%left TIMES DIVIDE MOD POWER
%left SQRT SUM AVG ABS LOG

%start program
%type <Ast.program> program

%%

program:
   /* nothing */ { [] }
 | program functiondecl { $2 :: $1 }

functiondecl:
   functype ID COLON formals_opt LBRACE variabledecl_list stmt_list RBRACE
     { { ftype = $1;
			fname = $2;
	 formals = $4;
	 locals = List.rev $6;
	 body = List.rev $7 } }

functype:
		FUNC VOID  { VoidRet }
	| FUNC INT { IntRet }
	| FUNC ARRAY { ArrayRet }

formals_opt:
    /* nothing */ { [] }
  | formal_list   { List.rev $1 }

formaldecl:
		INT ID               { Int($2) }
	| ARRAY ID              { Array($2) }

formal_list:
    formaldecl                        { [$1] }
  | formal_list COMMA formaldecl   { $3 :: $1 }

variabledecl_list:
    /* nothing */    { [] }
  | variabledecl_list variabledecl { $2 :: $1 }

variabledecl:
    INT ID SEMI { IntVar($2) }
	| ARRAY ID LPAREN LITERAL RPAREN SEMI { ArrayVar($2, $4) } 

variablecall:
		ID               { IntVarCall($1) }
	| ID LPAREN expr RPAREN { ArrayVarCall($1, $3) } 

stmt_list:
    /* nothing */  { [] }
  | stmt_list stmt { $2 :: $1 }

stmt:
    expr SEMI { Expr($1) }
  | RETURN expr SEMI { Return($2) }
  | LBRACE stmt_list RBRACE { Block(List.rev $2) }
  | IF LBRACKET expr RBRACKET stmt %prec NOELSE { If($3, $5, Block([])) }
  | IF LBRACKET expr RBRACKET stmt ELSE stmt    { If($3, $5, $7) }
  | FOR LBRACKET expr_opt SEMI expr_opt SEMI expr_opt RBRACKET stmt
     { For($3, $5, $7, $9) }
  | WHILE LBRACKET expr RBRACKET stmt { While($3, $5) }
	| PRINT expr SEMI { Print($2) }

expr_opt:
    /* nothing */ { Noexpr }
  | expr          { $1 }

expr:
    LITERAL          { Literal($1) }
  | expr PLUS   expr { Binop($1, Add,   $3) }
  | expr MINUS  expr { Binop($1, Sub,   $3) }
  | expr TIMES  expr { Binop($1, Mult,  $3) }
	| expr MOD    expr { Binop($1, Mod,  $3) }
  | expr DIVIDE expr { Binop($1, Div,   $3) }
	| expr POWER  expr { Power($1, $3) }
  | expr EQ     expr { Binop($1, Equal, $3) }
  | expr NEQ    expr { Binop($1, Neq,   $3) }
  | expr LT     expr { Binop($1, Less,  $3) }
  | expr LEQ    expr { Binop($1, Leq,   $3) }
  | expr GT     expr { Binop($1, Greater,  $3) }
  | expr GEQ    expr { Binop($1, Geq,   $3) }
	| SQRT expr        { Uniop(Sqrt, $2) }
	| SUM expr         { Uniop(Sum, $2) }
	| AVG expr         { Uniop(Avg, $2) }
	| ABS expr         { Uniop(Abs, $2) }
	| LOG expr         { Uniop(Log, $2) }
	| variablecall     { Var($1) }
  | variablecall ASSIGN expr   { Assign($1, $3) }
  | LBRACKET ID COLON actuals_opt RBRACKET { Call($2, $4) }
  | LBRACKET expr RBRACKET { $2 }

actuals_opt:
    /* nothing */ { [] }
  | actuals_list  { List.rev $1 }

actuals_list:
    expr                    { [$1] }
  | actuals_list COMMA expr { $3 :: $1 }
